/*
 * Created on 05.03.2009 by Johannes Textor
 * This Code is licensed under the BSD license:
 * http://www.opensource.org/licenses/bsd-license.php
 */
package infoatools;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Toolkit;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

/**
 * @author Johannes Textor
 *
 * This is a simple window class that permits basic drawing
 * operations (point and line). The window cannot be resized
 * after it was created.
 *
 */

public class Window {
	private SimpleCanvas theCanvas;

	private Color currentColor;

	/**
	 * Constructor for a new window. Note: depending on the
	 * speed of your system, it might take a while until a
	 * newly created window is actually shown on the screen.
	 * Any drawing operations that are executed before the window
	 * is fully displayed might yield strange results. However,
	 * this can be
	 * @param width Width of the window in pixels
	 * @param height Height of the window in pixels
	 * @param title String to be displayed in the window's
	 * title bar
	 */
	public Window(int width, int height, String title) {
		theCanvas = new SimpleCanvas(width, height);
		theCanvas.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent evt) {
				evt.getWindow().dispose();
			}

			public void windowOpened(WindowEvent evt) {
				theCanvas.repaint();
			}

			public void windowActivated(WindowEvent evt) {
				theCanvas.repaint();
			}
		});
		try {
			Thread.sleep(500);
		} catch (InterruptedException e) {
			//
		}
		theCanvas.addMouseListener(theCanvas);
		theCanvas.setTitle(title);
		theCanvas.setVisible(true);
		theCanvas.setResizable(true);
		currentColor = Color.WHITE;
	}

	/**
	 * Draw a straight line.
	 * @param x1 X coordinate of starting point
	 * @param y1 Y coordinate of starting point
	 * @param x2 X coordinate of endpoint
	 * @param y2 Y coordinate of endpoint
	 */
	public void drawLine(int x1, int y1, int x2, int y2) {
		Graphics g = theCanvas.getImage().getGraphics();
		g.setColor(currentColor);
		g.drawLine(x1, y1, x2, y2);
		theCanvas.repaint();
	}

	/**
	 * Draw a filled rectangle.
	 * @param x1 X coordinate of top left point
	 * @param y1 Y coordinate of top left point
	 * @param width Width of the rectangle
	 * @param height Height of the rectangle
	 */
	public void drawRect(int x1, int y1, int width, int height) {
		Graphics g = theCanvas.getImage().getGraphics();
		g.setColor(currentColor);
		g.fillRect(x1, y1, width, height);
		theCanvas.repaint();
	}

	/**
	 * Close the window.
	 * Your java program will continue to run until
	 * after the last window has been closed either
	 * manually by the user or by calling close().
	 */
	public void close() {
		theCanvas.dispose();
	}

	/** Set Color to be used by subsequent draw operations.
	 *
	 * @param r Red part of the color (0..255)
	 * @param g Green part of the color (0..255)
	 * @param b Blue part of the color (0..255)
	 */
	public void setColor(int r, int g, int b) {
		currentColor = new Color(r, g, b);
	}

	public void drawImage(Image i, int x, int y) {
		Graphics g = theCanvas.getImage().getGraphics();
		g.drawImage(i, x, y, null);
	}

	/**
	 * Draw an image in one of the formats supported by Java to 
	 * the canvas.
	 * @param filename Filename or URL of the image 
	 * @param x X coordinate
	 * @param y Y coordinate
	 */
	public void drawImage(String filename, int x, int y) {
		Image bild = Toolkit.getDefaultToolkit().getImage(filename);
		MediaTracker tracker = new MediaTracker(theCanvas);
		tracker.addImage(bild, 0);
		try {
			tracker.waitForID(0);
		} catch (InterruptedException e) {
		}
		drawImage(bild, x, y);
		theCanvas.repaint();
	}
	
	/**
	 * Waits until the user clicks somewhere in the window.
	 * @return The coordinates of the mouse click.
	 */
	public int[] waitForClick(){
		return theCanvas.waitForClick();
	}
	
	/**
	 * Draws text into the window.
	 * @param text The text to draw.
	 * @param x The x coordinate at which the text is drawn.
	 * @param y The y coordinate at which the text is drawn.
	 */
	public void drawString( String text, int x, int y ){
		theCanvas.getImage().getGraphics().drawString(text, x, y);
		theCanvas.repaint();
	}
}
